// HyperMD, copyright (c) by laobubu
// Distributed under an MIT license: http://laobubu.net/HyperMD/LICENSE
//
// DESCRIPTION: Auto show/hide markdown tokens like `##` or `*`
//
// Only works with `hypermd` mode, require special CSS rules
//
define(["require", "exports", "codemirror", "../core", "../core/cm_utils", "../core/line-spans"], function (require, exports, CodeMirror, core_1, cm_utils_1, line_spans_1) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    exports.getAddon = exports.HideToken = exports.suggestedOption = exports.defaultOption = void 0;
    var DEBUG = false;
    //#region Internal Function...
    /** check if has the class and remove it */
    function rmClass(el, className) {
        var c = ' ' + el.className + ' ', cnp = ' ' + className + ' ';
        if (c.indexOf(cnp) === -1)
            return false;
        el.className = c.replace(cnp, '').trim();
        return true;
    }
    /** check if NOT has the class and add it */
    function addClass(el, className) {
        var c = ' ' + el.className + ' ', cnp = ' ' + className + ' ';
        if (c.indexOf(cnp) !== -1)
            return false;
        el.className = (el.className + ' ' + className);
        return true;
    }
    exports.defaultOption = {
        enabled: false,
        line: true,
        tokenTypes: "em|strong|strikethrough|code|linkText|task|customLink".split("|"),
    };
    exports.suggestedOption = {
        enabled: true, // we recommend lazy users to enable this fantastic addon!
    };
    core_1.suggestedEditorConfig.hmdHideToken = exports.suggestedOption;
    core_1.normalVisualConfig.hmdHideToken = false;
    CodeMirror.defineOption("hmdHideToken", exports.defaultOption, function (cm, newVal) {
        ///// convert newVal's type to `Partial<Options>`, if it is not.
        if (!newVal || typeof newVal === "boolean") {
            newVal = { enabled: !!newVal };
        }
        else if (typeof newVal === "string") {
            newVal = { enabled: true, tokenTypes: newVal.split("|") };
        }
        else if (newVal instanceof Array) {
            newVal = { enabled: true, tokenTypes: newVal };
        }
        ///// apply config and write new values into cm
        var inst = (0, exports.getAddon)(cm);
        for (var k in exports.defaultOption) {
            inst[k] = (k in newVal) ? newVal[k] : exports.defaultOption[k];
        }
    });
    //#endregion
    /********************************************************************************** */
    //#region Addon Class
    var hideClassName = "hmd-hidden-token";
    var lineInactiveClassName = "hmd-inactive-line";
    var HideToken = /** @class */ (function () {
        function HideToken(cm) {
            var _this = this;
            this.cm = cm;
            this.renderLineHandler = function (cm, line, el) {
                // TODO: if we procLine now, we can only get the outdated lineView, lineViewMeasure and lineViewMap. Calling procLine will be wasteful!
                var changed = _this.procLine(line, el);
                if (DEBUG)
                    console.log("renderLine return " + changed);
            };
            this.cursorActivityHandler = function (doc) {
                _this.update();
            };
            this.update = (0, core_1.debounce)(function () { return _this.updateImmediately(); }, 100);
            /** Current user's selections, in each line */
            this._rangesInLine = {};
            new core_1.FlipFlop(
            /* ON  */ function () {
                cm.on("cursorActivity", _this.cursorActivityHandler);
                cm.on("renderLine", _this.renderLineHandler);
                cm.on("update", _this.update);
                _this.update();
                cm.refresh();
            }, 
            /* OFF */ function () {
                cm.off("cursorActivity", _this.cursorActivityHandler);
                cm.off("renderLine", _this.renderLineHandler);
                cm.off("update", _this.update);
                _this.update.stop();
                cm.refresh();
            }).bind(this, "enabled", true);
        }
        /**
         * hide/show <span>s in one line, based on `this._rangesInLine`
         * @returns line changed or not
         */
        HideToken.prototype.procLine = function (line, pre) {
            var cm = this.cm;
            var lineNo = typeof line === 'number' ? line : line.lineNo();
            if (typeof line === 'number')
                line = cm.getLineHandle(line);
            var rangesInLine = this._rangesInLine[lineNo] || [];
            var lv = core_1.cm_internal.findViewForLine(cm, lineNo);
            if (!lv || lv.hidden || !lv.measure)
                return false;
            if (!pre)
                pre = lv.text;
            if (!pre)
                return false;
            if (DEBUG)
                if (!pre.isSameNode(lv.text))
                    console.warn("procLine got different node... " + lineNo);
            var mapInfo = core_1.cm_internal.mapFromLineView(lv, line, lineNo);
            var map = mapInfo.map;
            var nodeCount = map.length / 3;
            var changed = false;
            var timeoutId = -1; // Declare a variable to store the timeout ID
            // change line status
            if (rangesInLine.length === 0) { // inactiveLine
                if (addClass(pre, lineInactiveClassName))
                    changed = true;
            }
            else { // activeLine
                if (rmClass(pre, lineInactiveClassName))
                    changed = true;
            }
            handleInlineDropdown();
            handleInlineAiContinuation();
            // show or hide tokens
            /**
             * @returns if there are Span Nodes changed
             */
            function changeVisibilityForSpan(span, shallHideTokens, iNodeHint) {
                var changed = false;
                iNodeHint = iNodeHint || 0;
                // iterate the map
                for (var i = iNodeHint; i < nodeCount; i++) {
                    var begin = map[i * 3], end = map[i * 3 + 1];
                    var domNode = map[i * 3 + 2];
                    if (begin === span.head.start) {
                        // find the leading token!
                        if (/formatting-/.test(span.head.type) && domNode.nodeType === Node.TEXT_NODE) {
                            // if (DEBUG) console.log("DOMNODE", shallHideTokens, domNode, begin, span)
                            // good. this token can be changed
                            var domParent = domNode.parentElement;
                            if (shallHideTokens ? addClass(domParent, hideClassName) : rmClass(domParent, hideClassName)) {
                                // if (DEBUG) console.log("HEAD DOM CHANGED")
                                changed = true;
                            }
                        }
                        //FIXME: if leading formatting token is separated into two, the latter will not be hidden/shown!
                        // search for the tailing token
                        if (span.tail && /formatting-/.test(span.tail.type)) {
                            for (var j = i + 1; j < nodeCount; j++) {
                                var begin_1 = map[j * 3], end_1 = map[j * 3 + 1];
                                var domNode_1 = map[j * 3 + 2];
                                if (begin_1 == span.tail.start) {
                                    // if (DEBUG) console.log("TAIL DOM CHANGED", domNode)
                                    if (domNode_1.nodeType === Node.TEXT_NODE) {
                                        // good. this token can be changed
                                        var domParent = domNode_1.parentElement;
                                        if (shallHideTokens ? addClass(domParent, hideClassName) : rmClass(domParent, hideClassName)) {
                                            changed = true;
                                        }
                                    }
                                }
                                if (begin_1 >= span.tail.end)
                                    break;
                            }
                        }
                    }
                    // whoops, next time we can start searching since here
                    // return the hint value
                    if (begin >= span.begin)
                        break;
                }
                return changed;
            }
            // This function will return the id of the customLink token by
            // searching for the class that contains the id
            function getIdFromToken(token) {
                if (token && token.classList && token.classList.value) {
                    var idx = token.classList.value.indexOf("customlink-id-");
                    var buffer = ("customlink-id-").length;
                    if (idx != -1) {
                        idx += buffer;
                        var id = token.classList.value.slice(idx, idx + 7);
                        return id;
                    }
                }
                return "";
            }
            var spans = (0, line_spans_1.getLineSpanExtractor)(cm).extract(lineNo);
            function getCurrentLine() {
                var lines = document.getElementsByClassName("CodeMirror-code")[0];
                var lineNumber = 0;
                var ret = { line: null, lineNumber: -1 };
                if (!lines || !lines.children)
                    return;
                for (var i = 0; i < lines.children.length; i++) {
                    var line = lines.children[i];
                    lineNumber++;
                    if (!(line && line.childElementCount == 1 && line.firstChild)) {
                    }
                    else if (!line.classList.contains("hmd-inactive-line")) {
                        ret.line = line;
                        ret.lineNumber = lineNumber;
                        return ret;
                    }
                }
                return ret;
            }
            function handleAiContinuationVisibility(e) {
                if (timeoutId != -1)
                    clearTimeout(timeoutId);
                timeoutId = setTimeout(function () {
                    var aiWidget = document.getElementById("ai-widget");
                    if (!aiWidget)
                        var aiWidget = createAiWidget();
                    var currentLine = getCurrentLine();
                    timeoutId = -1;
                    if (aiWidget && currentLine && currentLine.line && currentLine.line.firstChild && currentLine.line.firstChild.textContent.startsWith("#")) {
                        aiWidget.style.visibility = "hidden";
                        return;
                    }
                    // If the current line is empty, the arrow should not show
                    if (currentLine && currentLine.line && currentLine.lineNumber != -1) {
                        var prevLine = cm.getLineHandle(currentLine.lineNumber - 1);
                        var lineText = "";
                        if (!prevLine)
                            return;
                        lineText = prevLine.text;
                        if (lineText.trim().length < 1)
                            return;
                    }
                    if (aiWidget) {
                        if (currentLine && currentLine.line && currentLine.lineNumber != -1) {
                            var presentation = currentLine.line.firstChild;
                            if (presentation && presentation.role == "presentation") {
                                presentation.appendChild(aiWidget);
                            }
                        }
                    }
                }, 1000);
            }
            function createAiWidget() {
                var aiWidget = document.createElement('div');
                var aiStubResponse = [
                    "The population of the Earth is about 8 billion people, but there are over 10 quintillion (that's 10,000,000,000,000,000,000) ants!",
                    "The human brain uses about 20% of the body's total energy.",
                    "Cats have 32 muscles in each ear, allowing them to rotate their ears 180 degrees!",
                    "Chocolate was once used as currency.",
                    "There are more stars in the universe than grains of sand on all the beaches on Earth.",
                    "The world's quietest room is located at Microsoft's headquarters and is so quiet you can hear your own heartbeat.",
                    "Ketchup was originally sold as a medicine.",
                    "The world's first computer programmer was a woman named Ada Lovelace.",
                    "A group of owls is called a parliament.",
                    "The population of the Earth weighs about the same as 150 billion blue whales.",
                    "Mitocondria are the powerhouse of the cell.",
                    "The shortest war in history was between Zanzibar and England in 1896. Zanzibar surrendered after 38 minutes.",
                    "The longest time between two twins being born is 87 days.",
                    "The first oranges weren't orange.",
                    "The unicorn is the national animal of Scotland.",
                    "The first computer mouse was made of wood.",
                    "The first video game was created in 1958.",
                    "The first computer virus was created in 1971.",
                    "The first webcam was created in 1991.",
                ];
                aiWidget.id = "ai-widget";
                aiWidget.onclick = function (e) {
                    var currentLine = getCurrentLine();
                    if (currentLine.line && currentLine.lineNumber != -1) {
                        var lineHandle = cm.getLineHandle(currentLine.lineNumber - 1);
                        var lineText = lineHandle.text;
                        var response = lineText[lineText - 1] === " " || lineText[lineText - 1] === "   " ? "" : " ";
                        response += aiStubResponse[Math.floor(Math.random() * aiStubResponse.length)];
                        cm.replaceRange(response, { line: currentLine.lineNumber - 1, ch: lineText.length }, { line: currentLine.lineNumber - 1, ch: lineText.length });
                    }
                };
                return aiWidget;
            }
            // This function will handle the inline AI continuation widget 
            function handleInlineAiContinuation() {
                var editor = document.getElementsByClassName("CodeMirror")[0];
                // Create the AI widget if it doesn't already exist, also add a timeout to hide the widget 
                // after a certain amount of time
                if (editor) {
                    var aiWidgetCreated = false;
                    for (var i = 0; i < editor.children.length; i++) {
                        if (editor.children[i].id == "ai-widget") {
                            aiWidgetCreated = true;
                        }
                    }
                    if (!aiWidgetCreated) {
                        document.addEventListener('keydown', function (e) {
                            handleAiContinuationVisibility(e);
                        });
                        document.addEventListener('click', function (e) {
                            handleAiContinuationVisibility(e);
                        });
                    }
                }
            }
            // This function will handle the dropdown for the customLink options
            function onDropdownOptionClick(e, line) {
                var beginTagElem = null;
                if (e.target && e.target.parentElement && e.target.parentElement.parentElement) {
                    beginTagElem = e.target.parentElement.parentElement;
                }
                if (beginTagElem && beginTagElem.classList.value.includes("hmd-customlink-begin")) {
                    var id = getIdFromToken(beginTagElem);
                    var range = { from: -1, to: -1 };
                    var beginTagPos = 0;
                    for (var i = 0; i < line.children.length; i++) {
                        var elem = line.children[i];
                        if (elem && elem.classList && elem.classList.value.includes("hmd-customlink-begin")) {
                            beginTagPos++;
                            if (getIdFromToken(elem) == id) {
                                break;
                            }
                        }
                    }
                    for (var i = 0; i < spans.length && beginTagPos > 0; i++) {
                        var span = spans[i];
                        if (span && span.head.type.includes("hmd-customlink-begin")) {
                            beginTagPos--;
                            if (beginTagPos == 0) {
                                range.from = span.begin;
                                range.to = span.end;
                                if (range.from != -1 && range.to != -1)
                                    cm.replaceRange("[[".concat(e.target.textContent, "]]"), { line: lineNo, ch: range.from }, { line: lineNo, ch: range.to });
                                beginTagElem.innerHTML = '';
                                return;
                            }
                        }
                    }
                }
            }
            // This function will handle the dropdown for the customLink tokens
            // It will add the dropdown to the beginning of the customLink token
            function handleInlineDropdown() {
                var lines = document.getElementsByClassName("CodeMirror-code");
                var stubOptions = window.inlineDropdownOptions || [];
                if (lines && lines[0] && lines[0].children) {
                    var line = lines[0].children[lineNo];
                    if (line && line.children && line.firstChild) {
                        line = line.firstChild;
                        for (var i = 0; i < line.children.length; i++) {
                            var elem = line.children[i];
                            if (elem.classList.value.includes("cm-hmd-customlink-begin")) {
                                var beginTag = elem;
                                beginTag.classList.add("dropdown");
                                if (beginTag.children.length == 0) {
                                    var inlineDropdownContent = document.createElement('div');
                                    inlineDropdownContent.id = "active-inline-dropdown";
                                    inlineDropdownContent.classList.add("inline-dropdown-content");
                                    stubOptions.map(function (dropdownText) {
                                        var option = document.createElement('div');
                                        option.classList.add("dropdown-label", "dropdown-label-no-icon");
                                        option.innerHTML = dropdownText;
                                        option.onclick = function (e) {
                                            onDropdownOptionClick(e, line);
                                        };
                                        inlineDropdownContent.appendChild(option);
                                        beginTag.appendChild(inlineDropdownContent);
                                    });
                                }
                            }
                        }
                    }
                }
                var selection = document.getElementsByClassName("CodeMirror-selected");
                if (selection && selection[0]) {
                    var dropdowns = document.getElementsByClassName("inline-dropdown-content");
                    for (var i = 0; i < dropdowns.length; i++) {
                        var dropdown = dropdowns[i];
                        if (dropdown) {
                            dropdown.style.display = "none";
                        }
                    }
                    return;
                }
            }
            var iNodeHint = 0;
            for (var iSpan = 0; iSpan < spans.length; iSpan++) {
                var span = spans[iSpan];
                if (this.tokenTypes.indexOf(span.type) === -1)
                    continue; // not-interested span type
                /* TODO: Use AST, instead of crafted Position */
                var spanRange = [{ line: lineNo, ch: span.begin }, { line: lineNo, ch: span.end }];
                /* TODO: If use AST, compute `spanBeginCharInCurrentLine` in another way */
                var spanBeginCharInCurrentLine = span.begin;
                while (iNodeHint < nodeCount && map[iNodeHint * 3 + 1] < spanBeginCharInCurrentLine)
                    iNodeHint++;
                var shallHideTokens = true;
                for (var iLineRange = 0; iLineRange < rangesInLine.length; iLineRange++) {
                    var userRange = rangesInLine[iLineRange];
                    if ((0, cm_utils_1.rangesIntersect)(spanRange, userRange)) {
                        shallHideTokens = false;
                        break;
                    }
                }
                if (changeVisibilityForSpan(span, shallHideTokens, iNodeHint))
                    changed = true;
            }
            // finally clean the cache (if needed) and report the result
            if (changed) {
                // clean CodeMirror measure cache
                delete lv.measure.heights;
                lv.measure.cache = {};
            }
            return changed;
        };
        HideToken.prototype.updateImmediately = function () {
            var _this = this;
            this.update.stop();
            var cm = this.cm;
            var selections = cm.listSelections();
            var caretAtLines = {};
            var activedLines = {};
            var lastActivedLines = this._rangesInLine;
            // update this._activedLines and caretAtLines
            for (var _i = 0, selections_1 = selections; _i < selections_1.length; _i++) {
                var selection = selections_1[_i];
                var oRange = (0, cm_utils_1.orderedRange)(selection);
                var line0 = oRange[0].line, line1 = oRange[1].line;
                caretAtLines[line0] = caretAtLines[line1] = true;
                for (var line = line0; line <= line1; line++) {
                    if (!activedLines[line])
                        activedLines[line] = [oRange];
                    else
                        activedLines[line].push(oRange);
                }
            }
            this._rangesInLine = activedLines;
            if (DEBUG)
                console.log("======= OP START " + Object.keys(activedLines));
            cm.operation(function () {
                // adding "inactive" class
                for (var line in lastActivedLines) {
                    if (activedLines[line])
                        continue; // line is still active. do nothing
                    _this.procLine(~~line); // or, try adding "inactive" class to the <pre>s
                }
                var caretLineChanged = false;
                // process active lines
                for (var line in activedLines) {
                    var lineChanged = _this.procLine(~~line);
                    if (lineChanged && caretAtLines[line])
                        caretLineChanged = true;
                }
                // refresh cursor position if needed
                if (caretLineChanged) {
                    if (DEBUG)
                        console.log("caretLineChanged");
                    cm.refresh();
                    // legacy unstable way to update display and caret position:
                    // updateCursorDisplay(cm, true)
                    // if (cm.hmd.TableAlign && cm.hmd.TableAlign.enabled) cm.hmd.TableAlign.updateStyle()
                }
            });
            if (DEBUG)
                console.log("======= OP END ");
        };
        return HideToken;
    }());
    exports.HideToken = HideToken;
    //#endregion
    /** ADDON GETTER (Singleton Pattern): a editor can have only one HideToken instance */
    exports.getAddon = core_1.Addon.Getter("HideToken", HideToken, exports.defaultOption /** if has options */);
});
